/*
 * Copyright (c) 2004-2007 The Trustees of Indiana University and Indiana
 *                         University Research and Technology
 *                         Corporation.  All rights reserved.
 * Copyright (c) 2004-2006 The University of Tennessee and The University
 *                         of Tennessee Research Foundation.  All rights
 *                         reserved.
 * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
 *                         University of Stuttgart.  All rights reserved.
 * Copyright (c) 2004-2005 The Regents of the University of California.
 *                         All rights reserved.
 * Copyright (c) 2010      Cisco Systems, Inc. All rights reserved.
 * $COPYRIGHT$
 * 
 * Additional copyrights may follow
 * 
 * $HEADER$
 */
#include "ompi_config.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stddef.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif  /* HAVE_UNISTD_H */
#ifdef HAVE_STRING_H
#include <string.h>
#endif  /* HAVE_STRING_H */
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif  /* HAVE_SYS_TYPES_H */
#include <sys/mman.h>
//TODO: this will be in linux devkit

#if OMPI_BTL_HAVE_SRIO
#include <linux/rio_mport_cdev.h>
#include <linux/rio_regs.h>
#endif

#include "ompi/constants.h"
#include "opal/runtime/opal.h"
#include "opal/mca/event/event.h"
#include "opal/mca/base/mca_base_param.h"
#include "btl_srio.h"
#include "btl_srio_frag.h"
#include "btl_srio_fifo.h"
#include "pktfwdK2H/mpi_interfaces.h"
#include <ti/cmem.h>



/* CSL SRIO Functional Layer */
#include <ti/csl/tistdtypes.h>
#include <ti/csl/csl_types.h>
#include <ti/csl/cslr_device.h>
#include <ti/csl/csl_srio.h>
#include <ti/csl/csl_srioAux.h>
#include <ti/csl/csl_srioAuxPhyLayer.h>

#include <ti/csl/csl_serdes.h>
#include <ti/csl/csl_serdes_srio.h>

static int mca_btl_srio_component_progress (void);
static int srio_register(void);

void mpi_srioLog (char *report_string)
{
   opal_output_verbose(20, mca_btl_base_output, "SrioLog:%s", report_string);
}
void mpi_srioLogCRITICAL (char *report_string)
{
   opal_output_verbose(0, mca_btl_base_output, "SrioLogCRITICAL:%s", report_string);
}
#if 0
void srio_regs_custom(void *srio_base_virtual) {
 CSL_SrioHandle hSrio;
  uint32_t val,config_val,mask;
  
  hSrio=(CSL_SrioHandle )srio_base_virtual;
   
// 2)	Prescalar_select (bits 7-4) of RIO_PER_SET_CNTL = 0b1111, RIO_SP_RT_CTL to FFFFFFh.
   val=hSrio->RIO_PER_SET_CNTL;
   val = val & 0xFFFFFF0F;
   val = val | 0x000000F0;
  hSrio->RIO_PER_SET_CNTL=val;

  val = 0x00FFFFFF;
  hSrio->RIO_SP_RT_CTL=val;

// 3)	Set RIO_LSU_SETUP_REG1 (bits 9:8) = 0b00

   val=hSrio->RIO_LSU_SETUP_REG1;
//   val = val & 0xFFFFFCFF;
   val = val | 0x3 << 8;
   hSrio->RIO_LSU_SETUP_REG1=val;

// 4)	Please make sure port_writes are disabled at initialization(RIO_EM_DEV_PW_EN bit 0 and RIO_ERR_EN = 0x00000000 and RIO_SPn_RATE_EN = 0x00000000)

   hSrio->RIO_EM_DEV_PW_EN=hSrio->RIO_EM_DEV_PW_EN & 0xFFFFFFFE;
   hSrio->RIO_ERR_EN=0;
   hSrio->RIO_SP_ERR[0].RIO_SP_RATE_EN=0;
   hSrio->RIO_SP_ERR[1].RIO_SP_RATE_EN=0;
   hSrio->RIO_SP_ERR[2].RIO_SP_RATE_EN=0;
   hSrio->RIO_SP_ERR[3].RIO_SP_RATE_EN=0;

/*

In the rio_per_set_cntl register (0x0014).

Promote_Dis (Bit 21) = 0b1
TX_PRI2_WM (Bits 20-18) = 0b010
TX_PRI1_WM (Bits 17-15) = 0b011
TX_PRI0_WM (Bits 14-12) = 0b100
CRF_Credit (Bit 8) = 0b1
*/
/*
   val=hSrio->RIO_PER_SET_CNTL; 
   mask = ~ ((1<<21) |   (7 <<18) |   (7 <<15) |    (7 <<12) |   (1 <<8) )  ;
   config_val =         (1<<21) |   (2 <<18) |   (3 <<15) |    (4 <<12) |   (1 <<8) ;
  
   val = val & mask;
   val = val | config_val; 
   hSrio->RIO_PER_SET_CNTL=val;
  printf("\n MACHU DEBUG: RIO_PER_SET_CNTL = %x",hSrio->RIO_PER_SET_CNTL);
*/
  return;
}
#endif
/*
 * Shared Memory (SELF) component instance. 
 */

mca_btl_srio_component_t mca_btl_srio_component = {
    {  /* super is being filled in */
        /* First, the mca_base_component_t struct containing meta information
          about the component itself */
        {
            MCA_BTL_BASE_VERSION_2_0_0,

            "srio", /* MCA component name */
            OMPI_MAJOR_VERSION,  /* MCA component major version */
            OMPI_MINOR_VERSION,  /* MCA component minor version */
            OMPI_RELEASE_VERSION,  /* MCA component release version */
            mca_btl_srio_component_open,  /* component open */
            mca_btl_srio_component_close,  /* component close */
            NULL,
            srio_register,
        },
        {
            /* The component is checkpoint ready */
            MCA_BASE_METADATA_PARAM_CHECKPOINT
        },

        mca_btl_srio_component_init, 
        mca_btl_srio_component_progress,
    }  /* end super */
};
/*************************************************************************************************
 *******************************  Map SRIO Regs  *************************************************
 *************************************************************************************************/
static unsigned int srio_mmap_uio(unsigned int addr, unsigned int size,int uio_mem_fd)
{
        unsigned int virt_addr;
        unsigned int page_size;
        page_size = sysconf(_SC_PAGE_SIZE);
        if (size%page_size)
        {
                printf("Size does not align with page size. Size given: %d\n", size);
                return 0;
        }
        if ((unsigned int)addr % page_size)
        {
                printf("Address does not align with page size. Address given: 0x%08x\n", (unsigned int) addr);
                return 0;
        }
        virt_addr = (unsigned int) mmap(0, size, (PROT_READ|PROT_WRITE), MAP_SHARED, uio_mem_fd, (off_t)addr);
        if (virt_addr == -1)
        {
                printf("mmap failed!\n");
                return 0;
        }
        return virt_addr;
}
/***************************************************************************************************/
static char* mca_btl_srio_param_register_string(
    const char* param_name,
    const char* default_value)
{
    char *param_value;
    int id = mca_base_param_register_string("btl","srio",param_name,NULL,default_value);
    mca_base_param_lookup_string(id, &param_value);
    return param_value;
}

static int mca_btl_srio_param_register_int(
    const char* param_name,
    int default_value)
{
    int id = mca_base_param_register_int("btl","srio",param_name,NULL,default_value);
    int param_value = default_value;
    mca_base_param_lookup_int(id,&param_value);
    return param_value;
}
/**************************************************************************************/
static int srio_register(void)
{
    int i;
    FILE *node_id_detect, *cart_id_detect, *fd_rescan;
    unsigned char node_id_buff[4], cart_id_buff[4];
#if OMPI_BTL_HAVE_SRIO
    /*********************************************************************/
    /** Find location in topology: if this is EVM, default to 1 **/
    /*********************************************************************/
    node_id_detect = fopen("/proc/device-tree/sl1500/node/id", "rb");
    if(node_id_detect)
    {
       memset(node_id_buff, 0, 4);
       fread(node_id_buff, 1, 3, node_id_detect);
       mca_btl_srio_component.node_id = strtoul(node_id_buff, (char **)NULL, 10);
       fclose (node_id_detect);
    } else mca_btl_srio_component.node_id = 0;
    /** Find location in topology: if this is EVM, default to 1 **/
    cart_id_detect = fopen("/proc/device-tree/sl1500/cartridge/id", "rb");
    if(cart_id_detect)
    {
       memset(cart_id_buff, 0, 4);
       fread(cart_id_buff, 1, 3, cart_id_detect);
       mca_btl_srio_component.cartridge_id = strtoul(cart_id_buff, (char **)NULL, 10);
       fclose (cart_id_detect);
    } else mca_btl_srio_component.cartridge_id = 0;
    mca_btl_srio_component.srio_id = 1;
    mca_btl_srio_component.msmc_loc_addr =
        mca_btl_srio_param_register_int("msmc_loc_addr", BTL_SRIO_MSMC_ADDRESS);
    mca_btl_srio_component.cmem_cached  =
        mca_btl_srio_param_register_int("cmem_cached", BTL_SRIO_CMEM_CACHED);

    /* register SELF component parameters */
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_srio_component, "free_list_num",
                            "Number of fragments by default", false, false,
                            0, &mca_btl_srio_component.free_list_num );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_srio_component, "free_list_max",
                            "Maximum number of fragments", false, false,
                            -1, &mca_btl_srio_component.free_list_max );
    mca_base_param_reg_int( (mca_base_component_t*)&mca_btl_srio_component, "free_list_inc",
                            "Increment by this number of fragments", false, false,
                            32, &mca_btl_srio_component.free_list_inc );

    mca_btl_srio_component.frag_cnt[0] = mca_btl_srio_component.frag_cnt[1] = 0;

    mca_btl_srio_component.hostnameList = SrioHostnameListCreate (BTL_SRIO_PHYS_HOSTS);
    if(mca_btl_srio_component.hostnameList == NULL) {
       opal_output_verbose(10, mca_btl_base_output, "BTL_SRIO: host to physical location file %s cannot be opened\n", BTL_SRIO_PHYS_HOSTS);
       return OMPI_SUCCESS;
    }

    if(BTL_SRIO_EVM(mca_btl_srio_component)) {
      char my_hostname[80];
      gethostname(my_hostname, 80);
      if(SrioHostname2PhysID (mca_btl_srio_component.hostnameList, my_hostname, 
                              &mca_btl_srio_component.cartridge_id, &mca_btl_srio_component.node_id) < 0)
      {
        opal_output(0, "BTL_SRIO: Cannot find physical location of hostname %s, in list %s", my_hostname, BTL_SRIO_PHYS_HOSTS);
        return OMPI_ERR_OUT_OF_RESOURCE;
      }  
    } 
    strncpy (mca_btl_srio_component.routing_tbl_filename,
             mca_btl_srio_param_register_string("srio_routing_tbl", BTL_SRIO_ROUTING_TBL_FILE), 320);
    opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO: Routing table filename:%s cartridge_id=%d node_id=%d\n", 
      mca_btl_srio_component.routing_tbl_filename, mca_btl_srio_component.cartridge_id, mca_btl_srio_component.node_id);
    /** Import routing table **/
    mca_btl_srio_component.routingTbl = SrioRoutingStructuresCreate(mca_btl_srio_component.routing_tbl_filename); 
    if(mca_btl_srio_component.routingTbl == NULL) {
       opal_output_verbose(10, mca_btl_base_output, "BTL_SRIO: Routing table %s cannot be opened\n", mca_btl_srio_component.routing_tbl_filename);
       return OMPI_ERR_OUT_OF_RESOURCE;
    }

    mca_btl_srio_component.uio_mem_fd = open("/dev/srio", (O_RDWR | O_SYNC));

    if (mca_btl_srio_component.uio_mem_fd < 0) {
      opal_output(0, "UIO: Error opening SRIO device\n");
      return OMPI_ERR_OUT_OF_RESOURCE;
    }
    mca_btl_srio_component.srioregs_virtual     = (unsigned int *)srio_mmap_uio(0x2900000, 0x1c000, mca_btl_srio_component.uio_mem_fd);
    if (mca_btl_srio_component.srioregs_virtual == MAP_FAILED) 
    {
      opal_output(0, "UIO: Error mapping SRIO registers\n");
      return OMPI_ERR_OUT_OF_RESOURCE;
    }

    //srio_regs_custom(mca_btl_srio_component.srioregs_virtual); /* used by LSU only */

#endif
    return OMPI_SUCCESS;
}
/*
 *  Called by MCA framework to open the component, registers
 *  component parameters.
 */

int mca_btl_srio_component_open(void)
{
#if OMPI_BTL_HAVE_SRIO
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG: srio_component_open!!!\n");
    mca_btl_srio.super.btl_exclusivity = MCA_BTL_EXCLUSIVITY_HIGH;
    mca_btl_srio.super.btl_eager_limit = BTL_SRIO_EAGER_SIZE;
    mca_btl_srio.super.btl_rndv_eager_limit = BTL_SRIO_EAGER_SIZE;
    mca_btl_srio.super.btl_max_send_size = BTL_SRIO_MAX_SEND_SIZE;
    mca_btl_srio.super.btl_flags = MCA_BTL_FLAGS_SEND | MCA_BTL_FLAGS_SEND_INPLACE;
    mca_btl_srio.super.btl_seg_size = sizeof (mca_btl_base_segment_t);
    mca_btl_srio.super.btl_bandwidth = 10000;
    mca_btl_srio.super.btl_latency = 1;
    mca_btl_base_param_register(&mca_btl_srio_component.super.btl_version, &mca_btl_srio);
    /**************************************/
    mca_btl_srio_component.max_frag_size = mca_btl_srio.super.btl_max_send_size;
    mca_btl_srio_component.eager_limit = mca_btl_srio.super.btl_eager_limit;
    /**************************************/
    /* initialize objects */
    OBJ_CONSTRUCT(&mca_btl_srio_component.srio_lock, opal_mutex_t);
    OBJ_CONSTRUCT(&mca_btl_srio_component.srio_frags_eager, ompi_free_list_t);
    OBJ_CONSTRUCT(&mca_btl_srio_component.srio_frags_max, ompi_free_list_t);

//    if(CMEM_init() != 0) {
//      opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO: CMEM_init not available!");
//    }
#endif
    return OMPI_SUCCESS;
}
/********************************************************************************************************************/
/*
 * component cleanup - sanity checking of queue lengths
 */

int mca_btl_srio_component_close(void)
{
#if OMPI_BTL_HAVE_SRIO
    opal_output_verbose(20, mca_btl_base_output, "\nTIMPIDBG: srio_component_close!!!\n");
    OBJ_DESTRUCT(&mca_btl_srio_component.srio_lock);
    OBJ_DESTRUCT(&mca_btl_srio_component.srio_frags_eager);
    OBJ_DESTRUCT(&mca_btl_srio_component.srio_frags_max);
#endif
    return OMPI_SUCCESS;
}
/********************************************************************************************************************/
/*
 *  SRIO component initialization
 */
mca_btl_base_module_t** mca_btl_srio_component_init( int *num_btls, 
                                                     bool enable_progress_threads,
                                                     bool enable_mpi_threads )
{
    mca_btl_base_module_t **btls = NULL;
    char *ptr_hostname;
    int rc, i, tot_ports;
#if OMPI_BTL_HAVE_SRIO
    /** mpm_transport init **/
    mpm_transport_open_t ocfg = {
      .open_mode      = (O_SYNC|O_RDWR),
      .msec_timeout   = 2000
    };

    mpm_transport_mmap_t mcfg = {
      .mmap_prot      = (PROT_READ|PROT_WRITE),
      .mmap_flags     = MAP_SHARED,
    };
    opal_output_verbose(0, mca_btl_base_output, "TIMPIDBG: srio_component_init!!! (%d)", mca_btl_srio_component.node_id);

/*******/
#if 0
{
extern int TYPE11_main(int argc, char *argv[]);
char *list_of_arguments[3][5] = { 
   { "TYPE11_main", "c4", "sender", "c4n2", NULL },
   { "TYPE11_main", "c4", "sender", "c4n1", NULL },
   { "TYPE11_main", "c4", "router",   NULL, NULL }
};

  TYPE11_main (4, list_of_arguments[mca_btl_srio_component.node_id - 1]);
  opal_output_verbose(0, mca_btl_base_output, "TIMPIDBG: TYPE11 done !!!");
  sleep(3);
  exit(0);
}
#else
  
  if( ( mca_btl_srio_component.type11_drv = Srio_Type11_init (mca_btl_srio_component.routingTbl,
                                                              mca_btl_srio_component.cartridge_id, 
                                                              mca_btl_srio_component.node_id)) == NULL)
  {
    opal_output(0, "BTL_SRIO: Type11_init failed\n");
    return NULL;
  }
#endif
/*******/
    /* allocate the Shared Memory PTL */
    *num_btls = 1;
    btls = (mca_btl_base_module_t**)malloc((*num_btls)*sizeof(mca_btl_base_module_t*));
    if (NULL == btls) {
        return NULL;
    }
    if(!BTL_SRIO_EVM(mca_btl_srio_component)) 
    {
      if(mca_btl_srio_component.routingTbl == NULL) {
        opal_output(0, "BTL_SRIO: Routing table is not valid\n");
        return NULL;
      }
    } else {
      opal_output_verbose(10, mca_btl_base_output, "BTL_SRIO: Host to physical location file is not valid\n");
      return NULL;
    }
    mca_btl_srio_component.num_srio_btls = 1;
    mca_btl_srio_component.srio_btls     = btls;

    opal_output_verbose(100, mca_btl_base_output, "BTL_SRIO: free_list_num=%d free_list_max=%d free_list_inc=%d",
                        mca_btl_srio_component.free_list_num, mca_btl_srio_component.free_list_max, 
                        mca_btl_srio_component.free_list_inc);

    /* initialize free lists */
    rc = ompi_free_list_init_ex_new (&mca_btl_srio_component.srio_frags_eager,
                                     sizeof(mca_btl_srio_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_srio_frag1_t),
                                     mca_btl_srio.super.btl_eager_limit,
                                     opal_cache_line_size,
                                     mca_btl_srio_component.free_list_num,
                                     mca_btl_srio_component.free_list_max,
                                     mca_btl_srio_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_srio_frag_init,
                                     (void *) &mca_btl_srio);

    if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
        opal_output(0, "error creating srio fragment1 free list");
        return rc;
    }

    rc = ompi_free_list_init_ex_new (&mca_btl_srio_component.srio_frags_max,
                                     sizeof(mca_btl_srio_frag1_t),
                                     opal_cache_line_size, OBJ_CLASS(mca_btl_srio_frag1_t),
                                     mca_btl_srio.super.btl_max_send_size,
                                     opal_cache_line_size,
                                     mca_btl_srio_component.free_list_num,
                                     mca_btl_srio_component.free_list_max,
                                     mca_btl_srio_component.free_list_inc,
                                     NULL, (ompi_free_list_item_init_fn_t) mca_btl_srio_frag_init,
                                     (void *) &mca_btl_srio);

   if (OPAL_UNLIKELY(OMPI_SUCCESS != rc)) {
      opal_output(0, "error creating srio fragment2 free list");
      return rc;
   }
   /* get pointer to the btls */
   btls[0] = (mca_btl_base_module_t *)(&mca_btl_srio);
   /************************************************************************************************************************/
   /** SHARED MEMORY ACCESS: one MSMC region is mapped for both interfaces, and two DDR regions, one per HYPLNK interface **/
   /************************************************************************************************************************/
   /** Open local shared memory interface **/
   mca_btl_srio_component.mpm_transport_shmem = (void *)mpm_transport_open("arm-local-shmem", &ocfg);
   if (!mca_btl_srio_component.mpm_transport_shmem) {
      opal_output(0, "BTL SRIO shmem open failed");
      return NULL;
   }
   opal_output_verbose(20, mca_btl_base_output, "BTL SRIO shmem open successfull!!");
   {
     CMEM_AllocParams  alloc_paramsHeap;
     int32_t  num_Blocks  = 0;
     uint64_t ph_addr;

     CMEM_getNumBlocks(&num_Blocks);
     if (num_Blocks < 3)
     {
        opal_output(0, "MPI is using last CMEM block. At least 3 are expected and only %d are detected.\nBTL_SRIO won't be used.", num_Blocks);
        return NULL;
     }
     /* Allocate memory for BTL_SRIO_MAX_PROCESS fragment buffers.              */
     /* Each buffer is used for fragment assembly from multiple Type11 messages */
     /* In addition, pool of buffers used by SRIO and QMSS is allocated in Srio_Type11_init() */
     alloc_paramsHeap.type      = CMEM_HEAP;
     alloc_paramsHeap.alignment = 4;
     alloc_paramsHeap.flags     = mca_btl_srio_component.cmem_cached ? CMEM_CACHED : CMEM_NONCACHED;
     mca_btl_srio_component.virt_cmem_addr = CMEM_alloc2( num_Blocks - 1, BTL_SRIO_MMAP_SIZE, &alloc_paramsHeap);
     if(mca_btl_srio_component.virt_cmem_addr) {
       ph_addr = CMEM_getPhys (mca_btl_srio_component.virt_cmem_addr);
       mca_btl_srio_component.ph_cmem_addr = (uint32_t)(ph_addr + 0x080000000LL);
       opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO: CMEM physAddr:%16llx (to %08lx) userAddr:%p", ph_addr, 
       mca_btl_srio_component.ph_cmem_addr, mca_btl_srio_component.virt_cmem_addr);
     } else {
       opal_output_verbose(20, mca_btl_base_output, "BTL_SRIO: ERROR: CMEM_alloc2(). Cannot allocated from block %d!", num_Blocks - 1);
       return NULL;
     }
   }
   /****-----------------------------------------------****/
   /** Map local msmc region to user space               **/
   /** MSMC area will be used for packet forwarding and  **/ 
   /** buffer descriptors                                **/
   /****-----------------------------------------------****/
   mca_btl_srio_component.win_msmc = mpm_transport_mmap((mpm_transport_h)mca_btl_srio_component.mpm_transport_shmem, 
                                                         mca_btl_srio_component.msmc_loc_addr, BTL_SRIO_MSMC_MMAP_SIZE, &mcfg);
   if(!mca_btl_srio_component.win_msmc)
   {
     opal_output(0, "BTL SRIO mmap_failed_win_msmc!");
     return NULL;
   }
   opal_output_verbose(20, mca_btl_base_output, "BTL SRIO shmem MSMC mmap successfull!!");
   memset (mca_btl_srio_component.win_msmc, 0, BTL_SRIO_MSMC_MMAP_SIZE);

   mca_btl_srio_component.win_ddr      = (void *)((char *)mca_btl_srio_component.virt_cmem_addr);
   mca_btl_srio_component.ddr_loc_addr = mca_btl_srio_component.ph_cmem_addr;
   opal_output_verbose(10, mca_btl_base_output, "BTL SRIO turned ON !!!");
   return btls;
#else
   opal_output(0, "SRIO support not built in!!!\n");
   return NULL;
#endif
}
/*******************************************************************************************/
void mca_btl_srio_one_fragment_received (mca_btl_srio_hdr_t *hdr)
{
    mca_btl_srio_frag_t Frag;
    mca_btl_active_message_callback_t* reg;
//#ifdef BTL_SRIO_VERBOSE
    {
     char dbg_out[64];
     char *hdrc = (char *)hdr;
     sprintf (dbg_out, "TIMPIDBG-HDRfrag[%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x]",
                   hdrc[0], hdrc[1], hdrc[2], hdrc[3], hdrc[4], hdrc[5], hdrc[6], hdrc[7],hdrc[8], hdrc[9], hdrc[10], hdrc[11]);
     opal_output_verbose(20, mca_btl_base_output, "%s", dbg_out);
    }
// #endif
    memset (&Frag, 0, sizeof(mca_btl_srio_frag_t));
    /* recv upcall */
    reg = mca_btl_base_active_message_trigger + hdr->tag;
    Frag.base.des_dst = &Frag.segment;
    Frag.segment.seg_addr.pval = ((char *)hdr) + sizeof(mca_btl_srio_hdr_t);

    memcpy (&Frag.segment.seg_len, &hdr->len, sizeof(size_t));
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG:SRIO component_progress FRAG RCV 1 (len=%lld) end", Frag.segment.seg_len);
    Frag.base.des_dst_cnt = 1;
    reg->cbfunc(&mca_btl_srio.super, hdr->tag, &(Frag.base), reg->cbdata);
    opal_output_verbose(20, mca_btl_base_output, "TIMPIDBG:SRIO component_progress FRAG RCV 2 end");
}
/*******************************************/
/* Scan through all endpoints              */
/*******************************************/
void *btl_srio_endpoint_lookup (int srio_src_id)
{
  int i;

  for (i = 0; i < mca_btl_srio_component.endpoint_count; i ++)
  {
    if(mca_btl_srio_component.srio_endpoint[i].srio_id == srio_src_id) {
      if(srio_src_id == mca_btl_srio_component.srio_id) return NULL; // Not allowed to get looped packets!
      return &mca_btl_srio_component.srio_endpoint[i];
    }
  }
  return NULL;
}
/*******************************************************************************************/
extern uint64_t get_time_us(void);
extern void printSrioRtrStats(void);
extern void djdbg_update_queues(void);
extern void djdbg_report_queues(int tag_id);
volatile uint32_t djdbg_arr[64] = { 0 };
volatile int djdbg_dlab = 0;
volatile int report_counter  = 0;
static int mca_btl_srio_component_progress (void)
{
    mca_btl_srio_hdr_t *hdr;
    int peer_rank, rc = 0, nevents = 0;
    int j, ret_val;
    int ack_frag = 0;
    unsigned char *ptr, *frag_ptr, *ptr_start;
    void *cppi_desc;
    int  buff_len, curr_queue_depth;
    unsigned int seqn, max_seqn, src_id, dest_id;
    btl_srio_endpoint_t *srio_endpoint;
    uint64_t time_now;
    int var_verbosity = 0;
    static int output_throttle = 0;
    static int max_queue_depth = 0;
    static uint64_t time_start = 0;

djdbg_dlab = 101;
     output_throttle ++;
     if((output_throttle & 0x00000fff) == 0)
     {
       if(time_start == 0) time_now = time_start = get_time_us();
       else time_now = get_time_us();
       if((time_now - time_start) > 10000000) {
         opal_output_verbose(0,  mca_btl_base_output, "[%3d] PDSP stats (%lld)", report_counter, time_now);
         djdbg_update_queues();
         djdbg_report_queues(0);
//       printSrioRtrStats ();
         max_queue_depth = 0;
         report_counter ++;
         if(report_counter > 4) sleep (10);
         time_start = get_time_us();; //reset the timer
       }
     }
djdbg_dlab = 1;
#if OMPI_BTL_HAVE_SRIO
    for(j = 0; j < 10; j ++)
    { // Scan all SRIO ports
djdbg_dlab = 2;
//  probe_pt_start(0);
      cppi_desc = Srio_Type11_AcquireRX(mca_btl_srio_component.type11_drv, &ptr, &buff_len); /* returns cppi_desc */
//  probe_pt_end(0);
djdbg_dlab = 3;
      if(!cppi_desc) continue;
djdbg_dlab = 4;
      ptr_start = ptr;
//  probe_pt_start(1);
      btl_srio_atomic_dsb(); /* << Memory barrier, make sure all the CPU transfers are done by this point, data are in CACHE */
// if(mca_btl_srio_component.cmem_cached)
 CMEM_cacheInv(ptr_start, 2048);
      //btl_srio_atomic_dsb(); /* << Memory barrier, make sure all the CPU transfers are done by this point, data are in DDR */
      /* Memory consistency */
      btl_srio_atomic_dmb(); /* << Memory barrier, make sure all the CPU transfers are done by this point, data are in CACHE */
//  probe_pt_end(1);
djdbg_dlab = 5;
      dest_id = (unsigned int)*ptr++; /* destination ID */
      src_id  = (unsigned int)*ptr++; /* source ID */
      seqn    = (unsigned int)*ptr++; /* 2K message sequence number */
      max_seqn =(unsigned int)*ptr++; /* total number of 2K messages */
#if 1
      if((buff_len != 2048) || (max_seqn > 32)) {
         opal_output_verbose(0,  mca_btl_base_output, "Type11 message sanity check failed... dropping! cppi %p ptr %p (seqn=%d max_seqn=%d src_id=%d dest_id=%d buff_len=%d)",
                             cppi_desc, ptr_start, seqn, max_seqn, src_id, dest_id, buff_len);
         opal_output_verbose(0, mca_btl_base_output, "Type11 message sanity check failed... dropping! cppi %08x descSize %d PSdata: srcid %d dstid %d mbox %d letter %d num_bytes %d ttType11 %d pri %d ccType11 %d raw:%08x %08x ", cppi_desc, 
            djdbg_arr[0], djdbg_arr[1], djdbg_arr[2], djdbg_arr[3], djdbg_arr[4], djdbg_arr[5], djdbg_arr[6],
            djdbg_arr[7], djdbg_arr[8], djdbg_arr[9], djdbg_arr[10], djdbg_arr[11], djdbg_arr[12]);
/*
djdbg_arr[1] = ptr_srioDrvInst->rxDescSize;
djdbg_arr[2] = srcID;
djdbg_arr[3] = dstID;
djdbg_arr[4] = mbox;
djdbg_arr[5] = letter;
djdbg_arr[6] = num_bytes;
djdbg_arr[8]  = ttType11;
djdbg_arr[9]  = pri;
djdbg_arr[10] = ccType11;
djdbg_arr[11] = srioInfo[0];
djdbg_arr[12] = srioInfo[1];
*/
         Srio_Type11_ReleaseRX(mca_btl_srio_component.type11_drv, cppi_desc);
         continue;
      }
#endif
      opal_output_verbose(10,  mca_btl_base_output, "Type11 message received! %p (seqn=%d max_seqn=%d src_id=%d dest_id=%d buff_len=%d)",
                          ptr_start, seqn, max_seqn, src_id, dest_id, buff_len);
      buff_len -= 4 * sizeof(unsigned char);
      if(dest_id != mca_btl_srio_component.srio_id) {
        extern void djdbg_report_psinfo(void *cppi_desc);
        opal_output_verbose(0,  mca_btl_base_output, "wrong routing in Type11 message (my dest_id is %d, received dest_id %d src_id %d, seqn=%d, max_seqn=%d)",
                                  mca_btl_srio_component.srio_id, dest_id, src_id, seqn, max_seqn);
        djdbg_report_psinfo(cppi_desc);

        Srio_Type11_ReleaseRX(mca_btl_srio_component.type11_drv, cppi_desc);
        continue;
      } 
      if(!(srio_endpoint = btl_srio_endpoint_lookup(src_id))) {
        opal_output_verbose(0,  mca_btl_base_output, "wrong source id in Type11 message (%d)", src_id);
        Srio_Type11_ReleaseRX(mca_btl_srio_component.type11_drv, cppi_desc);
        continue;
      }
      opal_output_verbose(10,  mca_btl_base_output, "Type11 message endpoint: srio_id %d idx %d prevseqn %d->seqn %d",
                               srio_endpoint->srio_id, srio_endpoint->idx, srio_endpoint->prev_seqn, seqn);
      if((srio_endpoint->prev_seqn + 1) != seqn) {
         opal_output_verbose(0,  mca_btl_base_output, "Type11 message sequence DISCONTINUITY (srio_id %d idx %d prevseqn %d->seqn %d max_seqn %d)",
                             srio_endpoint->srio_id, srio_endpoint->idx, srio_endpoint->prev_seqn, seqn, max_seqn);
djdbg_dlab = 6;
//         if(mca_btl_srio_component.cmem_cached) 
         CMEM_cacheInv(ptr_start, 2048);
         /* Recovery */
         dest_id = (unsigned int)ptr_start[0]; /* destination ID */
         src_id  = (unsigned int)ptr_start[1]; /* source ID */
         seqn    = (unsigned int)ptr_start[2]; /* 2K message sequence number */
         max_seqn =(unsigned int)ptr_start[3]; /* total number of 2K messages */
         opal_output_verbose(0,  mca_btl_base_output, "Type11 message sequence DISCONTINUITY after cacheInv, seqn=%d max_seqn=%d", 
                             ptr_start[2], ptr_start[3]);
djdbg_dlab = 7;
      }
      //btl_srio_dump_buffer("RXfragBEG", (unsigned int *)ptr, 0, 4);
      //btl_srio_dump_buffer("RXfragEND", (unsigned int *)ptr, 480, 4);
//  probe_pt_start(2);
      frag_ptr = srio_endpoint->virt_baseaddr + srio_endpoint->idx;
      memcpy (frag_ptr, ptr, buff_len);
      btl_srio_atomic_dsb(); /* << Memory barrier, make sure all the CPU transfers are done by this point */
      //btl_srio_atomic_dmb(); /* << Memory barrier, make sure all the CPU transfers are done by this point */
djdbg_dlab = 8;
      Srio_Type11_ReleaseRX(mca_btl_srio_component.type11_drv, cppi_desc);
//  probe_pt_end(2);
djdbg_dlab = 9;
      if(seqn >= max_seqn)
      { /* last segment ? */
        opal_output_verbose(10, mca_btl_base_output, "BTL_SRIO component_progress fragment, Type11 messages %d, from %d", 1 + seqn, src_id);
djdbg_dlab = 10;
        //btl_srio_dump_buffer("RXFINBEG", (unsigned int *)srio_endpoint->virt_baseaddr, 0, 8);
        //btl_srio_dump_buffer("RXFINEND", (unsigned int *)srio_endpoint->virt_baseaddr, 484, 8);
djdbg_dlab = 11;
        mca_btl_srio_one_fragment_received ((mca_btl_srio_hdr_t *)srio_endpoint->virt_baseaddr);
djdbg_dlab = 12;
        srio_endpoint->idx       = 0;
        srio_endpoint->prev_seqn = -1;
        nevents ++;
#if 1
        time_start = get_time_us(); //reset the timer
#endif
djdbg_dlab = 13;
        continue;
      }
djdbg_dlab = 15;
      srio_endpoint->idx += buff_len;
      srio_endpoint->prev_seqn = seqn;
      opal_output_verbose(10,  mca_btl_base_output, "Type11 message buffer RELEASED (srio_id %d idx %d seqn %d)",
                          srio_endpoint->srio_id, srio_endpoint->idx, seqn);
djdbg_dlab = 16;
    }
djdbg_dlab = 100;
#endif
    return 1;
}

/*** nothing past this point ***/
