/**
 *   @file  srio_test.h
 *
 *   @brief   
 *      This files contains the definitions and data structure used in the test code.
 *
 *  \par
 *  ============================================================================
 *  @n   (C) Copyright 2014, Texas Instruments, Inc.
 * 
 *  Redistribution and use in source and binary forms, with or without 
 *  modification, are permitted provided that the following conditions 
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the   
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  \par
*/

#ifndef __SRIO_TEST_H__
#define __SRIO_TEST_H__

#include "srioPlatCfg.h"
#include <ti/drv/qmss/qmss_srioRtr.h>
#include "rtr_sriodrv.h"

/************************ USER DEFINES ********************/

#define RM 1

#define NUMBER_OF_CORES             srio_EXAMPLE_NUM_CORES

#define SYSINIT                     0

/*   TERMINATION = 144
 *   Tx Host Descriptors = 4*4 ports  = 16
 *   Rx Host Descriptors = 32*4 ports = 128
 *   CREDITS     = 24
 *   Tx Credit Descriptors = 4 * 4 ports = 16 (4 Pools, pre-prepared with different settings(dest-id different) per port)
 *   Rx Credit Descriptors = 4 * 2 = 8 (2 per port, for ping pong)
 *   FORWARDING  = 32
 *   Tx Forwarding Descriptors = 4 * 4 ports = 16 (4 pools, pre-prepared with dest-id = neighbhouring node on this port )
 *   Rx Forwarding Descriptors = 4 * 4 ports = 16 (single pool)
 *   TOTAL = 200 descriptors (size is 64 bytes).
 */
/* All descriptors and forwarded data buffers, including credit data buffers are in MSMC */
/* Terminated data buffers (generated or consumed by A15), both TX and RX, are in DDR */
#define SRIO_MAX_MTU                2048 /* was 256 */
#define SRIO_MAX_CREDIT_MTU           16 /* */

#define NUM_TERMINATE_TX_DESCS (64*4) // was 32*4
#define NUM_TERMINATE_RX_DESCS (4096*4)  // was 32*4
#define NUM_TERMINATE_PACKETS_DDR 32768 //(NUM_TERMINATE_TX_DESCS + NUM_TERMINATE_RX_DESCS)// Originally (16+128), making it a power of 2 for QMSS insert mem region

#define NUM_FWD_POOL_PER_PORT       256
#define NUM_FWD_PACKETS_MSMC       (4*NUM_FWD_POOL_PER_PORT)
#define SRIO_CREDIT_SIZE_INCREMENT  128
#define PDSP_CREDIT_PERIOD      SRIO_CREDIT_SIZE_INCREMENT // (NUM_FWD_POOL_PER_PORT/2)   /* Credit is sent once every (NUM_FWD_POOL_PER_PORT/2) packets are arrived at the rx forward queue */
#define NUM_CREDIT_TX_DESCS 16
#define NUM_CREDIT_RX_DESCS 16

#define NUM_CREDIT_PACKETS_MSMC    (NUM_CREDIT_TX_DESCS + NUM_CREDIT_RX_DESCS)  //Originally (16+8), making it a power of 2 for QMSS insert mem region

#define NUM_HOST_DESC              (NUM_TERMINATE_PACKETS_DDR + NUM_FWD_PACKETS_MSMC + NUM_CREDIT_PACKETS_MSMC )//256 //(16+128+16+8+16+16)
#define SIZE_HOST_DESC             64

#define DDR_SIZE_DATA_BUFFER       (NUM_TERMINATE_PACKETS_DDR*SRIO_MAX_MTU) /* 2K data packet size */
#define MSMC_SIZE_DATA_BUFFER      (NUM_FWD_PACKETS_MSMC*SRIO_MAX_MTU) /*  2K data packet size */
#define MSMC_SIZE_CREDIT_BUFFER    (NUM_CREDIT_PACKETS_MSMC*SRIO_MAX_CREDIT_MTU)

/* For exp-1, making all queues of the same priority, and CRF bit = 0 for all */
#define SRIO_TX_DATA_QUEUE_PRIORITY 0
#define SRIO_TX_CREDIT_QUEUE_PRIORITY 0
#define SRIO_QUEUE_BASE		   672
/* In MSMC we need (16+16)x2KB for forwarding buffers and 200x64Bytes = 12.5KB for descriptors.  */
/* MSMC buffers are allocated using mpm_transport_mmap() API                                     */
/* In DDR we will need (16+128)x2KB=288KB for termination buffers. This will be allocated from CMEM */
/* 1. We have one memory area in MSCM for TxRx host descriptors */
/* 1a. We have 4 more memory areas in MSMC for TX credit descriptors */
/* 1b. We have one memory area in MSMC for RX credit descriptors */
/* 2. We have one memory area in MSMC for buffers used for forwarded traffic */
/* 3. We have one memory area in DDR for buffer used for termination */

extern Rm_ServiceHandle *rmClientServiceHandle;
extern uint8_t *host_region_MSMC; // descriptors for credits and forwarding */
extern uint8_t *host_region_DDR;  // descriptors for host terminated tx and/or rx */

extern Qmss_Result    qmssRegion_terminate;
extern Qmss_Result    qmssRegion_credit;
extern Qmss_Result    qmssRegion_forward;

extern Uint8   *host_region_terminate;
extern Uint8   *host_region_credit;
extern Uint8   *host_region_forward;
//extern uint8_t *dataBuff[NUMBER_OF_CORES];

extern uint8_t *dataBuff_DDR;   // host packets tx/rx.  TX ~128 (64K*2), RX 4*2K*1024 startIndex UNSPECIFIED (int preferred, but ext allowed)
extern uint8_t *dataBuff_MSMC;  // credit (64b, 4b used) 2*4*64 TX precooked in 4 pools, 1 or 4 RX pools total 2*4*64; startIndex for region INTERNAL

#define TEST_PDSP                   Qmss_PdspId_PDSP2

/* virtual core/task number */
extern uint32_t coreNum;

/* error counter */
extern uint32_t errorCount;


typedef enum {
  ROLE_SENDER=0,
  ROLE_RECEIVER=1,
  ROLE_ROUTER=2,
} Role_e;

Role_e  Role_global;
typedef struct {
  int remote_id;
  int connected_port;
  void *socket;
  int num_packets_sent;
  int total_num_packets_sent;
  int num_packets_received;
  int total_num_packets_received;
  struct timespec recv_start;
  struct timespec recv_finish;
  int sync_packet_received;
  volatile int print_send_stats;
  volatile int print_num_packets_sent;
  volatile int print_total_num_packets_sent;
  volatile int print_receive_stats;
  volatile int print_num_packets_received;
  volatile int print_total_num_packets_received;
} remoteInfoEntry_t;



#define MAX_REMOTE_INFO_ENTRIES 36

typedef struct {
  remoteInfoEntry_t entries[MAX_REMOTE_INFO_ENTRIES];
  int num_entries;
} remoteInfo_t;


typedef struct {
  /************ Terminiation Queues ******************/
  Qmss_QueueHnd  tx_terminate_queues[4]; /* To be used by A15 to write in to, during sock Send. PDSP will be popping from this queue.
                                          This will be assigned to the ptrSocket->txQueue.
                                          This will be the 'hostSrioTxQueue' in the firmware
                                          Return to Free Pool: txFree_terminate_queues[] */
  Qmss_QueueHnd  txFree_terminate_queues[4]; /* To be used by A15 to allocate from , during sock Send. PDSP will be popping from this queue,
                                              This will be assigned to the ptrSocket->txFreeQueue */

  Qmss_QueueHnd  rx_terminate_queue; /* This is the SrioDrv->Rx Completion Queue. Wiill not be used by PDSP. From the incoming packets, based on the letter value, the SRIO Hw will
                                          push it in to this queue. We need to make sure the RXU_ registers are mapped properly.
                                          Return to Free Pool: rxFree_terminate_queue[] */

  Qmss_QueueHnd  rxFree_terminate_queue; /* This is the SrioDrv->RxQueue.  Packets from the rx_terminate queue, once they are processed, will find itself on this queue */


  /************ Credit Queues ******************/

  Qmss_QueueHnd  tx_srio_credit_queues[4]; /* Completely used by PDSP. This will be the 'creditSrioTxQueue' in the PDSP firmwire .
                                         Return to Free Pool: txFree_credit_queues[]. This HQ queue will be 673+ (port*4)  */
  Qmss_QueueHnd  txFree_credit_queues[4]; /* PDSP will pop a descriptor from this queue and place in to tx_credit_queue.
                                           This queue will be the 'creditSourceQueue' inside the PDSP firmwire  */

  Qmss_QueueHnd  rx_credit_queues[4]; /* Completely used by PDSP */
  Qmss_QueueHnd  rxFree_credit_queues[4]; /* Completely used by PDSP. The descrpitors, once processed from the above rx_credit_queue, will be placed here */


  /************ TX SRIO HW Queues ******************/

  Qmss_QueueHnd  tx_srio_queues[4]; /* Completely used by PDSP, This is a SRIO HW Queue, starting at 672 on K2H . This will be the
                                        sourcePortSrioTxQueue[]. If the originating port doesn't matter, all values in sourcePortSrioTxQueue[]
                                        can have the same value.  The descriptor will be returned to either txFree_terminate_queues[] or
                                        rxFree_forward_queues[] depending up on where PDSP routes the descriptor it from.i.e either tx_terminate_queues (emanating from A15)
                                        or rx_completed_forward_queues[] (received by SRIO HW, meant to be forwarded) */


  /************ RX SRIO HW Queues ******************/
  Qmss_QueueHnd  rx_completed_forward_queues[4]; /* SRIO RXU will push entries in to this queue, by popping from the below rxFree_forward_queues queue.
                                                    . PDSP will now pop from rx_completed_forward_queues[] on to tx_srio_queue..
                                                    Link to Free Pool: rxFree_forward_queues[]  */
  Qmss_QueueHnd  rx_forward_return_queues[4]; /* Completely used by PDSP to pop from rx_completed_forward_queues(above) to this queue after processing */

  Qmss_QueueHnd  rxFree_forward_queues[4]; /* Completely used by PDSP, by popping free descriptors from this queue and pushing it on to the tx_forward_queues */

  Qmss_QueueHnd  debugQueues[4];

  Qmss_QueueHnd  errorQueue;
  Qmss_QueueHnd  emptyQueue;
  Qmss_QueueHnd  queueBase; /* BAse of the block queue consisting of TX Terminate, Rx FWD and RX Credit Queues, used for PDSP RTR Config */
  Cppi_Handle cppiHnd;/* CPPI handle for Packet DMA */

} SrioQueues_t;

/********************************************************************************************************/
/********************************************************************************************************/
/********************************************************************************************************/

typedef struct {
	uint8_t portToForward[256]; /* portToForward[#id] = the port to forward to, if the destination ID of the received packet is 'id' */
	uint8_t is_dest_field[256];  /* dest_field[id] = 0/1 , denotes if the 'id' is an adjacent node in the topology.
                               This info is used by the PDSP router to send Tx Credit request or not */
}Srio_PDSPRouterCfg_t ;

/* Function prototypes */
void example_main(Cppi_Handle cppiHnd);

typedef struct {
	Srio_SockHandle socket_per_port[4]; /* Socket for each SRIO port, for communicating with the immediate neighbour */
    int my_id;  /* My SRIO ID */
    int dest_ids_port[4]; /* Neighbour's SRIO id at the other end of my outgoing port */
    Srio_PDSPRouterCfg_t pdspRouterConfig;
    SrioQueues_t srioQueues;
    Srio_DrvHandle srioDrvHandle;
} Sriotype11APIsContext_t;


extern Sriotype11APIsContext_t Sriotype11APIsContext;


int multinode_test_main(void *args);


void fw_task_wait (void *handle);

void fw_osalInit();

void fw_osalshutdown() ;

void fw_osalEnterCS();

void fw_osalLeaveCS();

uint32_t fw_name_lookup (const char *name);

int setupTestFramework (Cppi_Handle *cppiHnd);

/* set name to value */
void fw_name_set (const char *name, uint32_t val);

/* delete name */
 void fw_name_del (const char *name);

 int32_t Osal_dataBufferInitMemory(uint32_t dataBufferSize);

/**
  Unit test of this type:
  Srio_Type11_init();
  Srio_Type11_setup();
  for (::) {
    Srio_Type11_AcquireTX(.., &ptr, ...);
    memset(ptr, ...);
    Srio_Type11_SendTX(..., ptr, ...);
    if(Srio_Type11_AcquireRX(...)) {
      printf ("\nContent of buffer...\");
      Srio_Type11_ReleaseRX(...);
    }
  }
  Srio_Type11_finalize();
**/
/**************************************************/

#endif /* __SRIO_TEST_H__ */

